# Terraform Test Framework
# https://github.com/tf2project/tf2project

from .local import TerraformStateLoader
from .object import TerraformObject, TerraformObjectParser
from .util import get_attr_name, get_output_type


class Terraform:
    def __init__(self, loader_instance=None):
        if loader_instance is None:
            self._loader_instance = TerraformStateLoader()
        else:
            self._loader_instance = loader_instance
        self._load_data()
        self._fetch_target_values()
        self._parse_modules(self, self._target_values["root_module"])
        self._parse_outputs()

    def get_loader_instance(self):
        return self._loader_instance

    def get_data(self):
        if hasattr(self, "_data") is False:
            raise Exception("Data is not loaded.")
        return self._data

    def _load_data(self):
        self._data = self._loader_instance.load()

    def _fetch_target_values(self):
        loader_type = self._loader_instance.get_loader_type()
        if loader_type == "planloader":
            self._target_values = self._data["planned_values"]
        elif loader_type == "stateloader":
            self._target_values = self._data["values"]
        else:
            raise Exception("Invalid terraform resource loader.")

    def _parse_modules(self, target_object, target_module):
        if "resources" in target_module:
            self._parse_resources(target_object, target_module["resources"])
        if "child_modules" in target_module:
            target_object.modules = TerraformObject()
            target_object.modules.__all__ = {}
            for module in target_module["child_modules"]:
                module_name = get_attr_name(module["address"].split(".")[-1])
                setattr(target_object.modules, module_name, TerraformObject())
                module_object = getattr(target_object.modules, module_name)
                target_object.modules.__all__[module_name] = module_object
                self._parse_modules(module_object, module)

    def _parse_resources(self, target_object, resources):
        for resource in resources:
            if resource["mode"] == "managed":
                if hasattr(target_object, "resources") is False:
                    target_object.resources = TerraformObject()
                    target_object.resources.__all__ = {}
                resource_mode_object = target_object.resources
            elif resource["mode"] == "data":
                if hasattr(target_object, "data") is False:
                    target_object.data = TerraformObject()
                    target_object.data.__all__ = {}
                resource_mode_object = target_object.data
            else:
                raise Exception("Invalid terraform resource mode.")
            if hasattr(resource_mode_object, resource["type"]) is False:
                setattr(
                    resource_mode_object,
                    resource["type"],
                    TerraformObject(),
                )
            resource_type_object = getattr(
                resource_mode_object,
                resource["type"],
            )
            if resource["type"] not in resource_mode_object.__all__:
                resource_mode_object.__all__[resource["type"]] = resource_type_object
                resource_type_object.__all__ = {}
            resource_name = get_attr_name(resource["name"])
            if "index" in resource:
                if hasattr(resource_type_object, resource_name) is False:
                    setattr(
                        resource_type_object,
                        resource_name,
                        TerraformObject(),
                    )
                resource_name_object = getattr(
                    resource_type_object,
                    resource_name,
                )
                if type(resource["index"]) is str:
                    if hasattr(resource_name_object, "instances") is False:
                        setattr(resource_name_object, "instances", {})
                    resource_name_object.instances.update(
                        {resource["index"]: TerraformObjectParser(resource)}
                    )
                elif type(resource["index"]) is int:
                    if hasattr(resource_name_object, "instances") is False:
                        setattr(resource_name_object, "instances", [])
                    resource_name_object.instances.append(
                        TerraformObjectParser(resource)
                    )
                else:
                    raise Exception("Invalid terraform resource index.")
            else:
                setattr(
                    resource_type_object,
                    resource_name,
                    TerraformObjectParser(resource),
                )
                resource_name_object = getattr(
                    resource_type_object,
                    resource_name,
                )
            if resource_name not in resource_type_object.__all__:
                resource_type_object.__all__[resource_name] = resource_name_object

    def _parse_outputs(self):
        if "outputs" not in self._target_values:
            return None
        self.outputs = TerraformObject()
        self.outputs.__all__ = {}
        for _key, value in self._target_values["outputs"].items():
            key = get_attr_name(_key)
            setattr(self.outputs, key, TerraformObject())
            target_object = getattr(self.outputs, key)
            self.outputs.__all__[key] = target_object
            target_object.sensitive = value["sensitive"]
            if "value" in value:
                target_object.type = get_output_type(value["value"])
                if type(value["value"]) is dict:
                    target_object.value = TerraformObjectParser(
                        data=value["value"], is_testable=False
                    )
                else:
                    target_object.value = value["value"]
            else:
                target_object.type = target_object.value = None



# Terraform Test Framework
# https://github.com/tf2project/tf2project

from subprocess import TimeoutExpired
from subprocess import run as run_subprocess

from .executor import Executor


class LocalCommandExecutor(Executor):
    def execute(self, command, *args, **kwargs):
        try:
            result = run_subprocess(
                command,
                shell=True,
                capture_output=True,
                *args,
                **kwargs,
            )
        except TimeoutExpired as e:
            timeout = True
        except:
            pass
        finally:
            if "result" in locals():
                self._is_complete = True
                self._set_result(
                    command=command,
                    rc=result.returncode,
                    stdout=result.stdout.decode("utf8"),
                    complete=self._is_complete,
                )
            else:
                self._is_complete = False
                self._set_result(
                    command=command,
                    timeout=True if "timeout" in locals() else False,
                    complete=self._is_complete,
                )



# Terraform Test Framework
# https://github.com/tf2project/tf2project

from .executor import *
from .terraform import *
from .test import *
from .tf2 import *

__version__ = "0.2.0"

__all__ = [
    "Executor",
    "LocalCommandExecutor",
    "Terraform",
    "TerraformLoader",
    "TerraformLocalFileLoader",
    "TerraformPlanLoader",
    "TerraformStateLoader",
    "Test",
    "Tf2",
]



# Terraform Test Framework
# https://github.com/tf2project/tf2project


class Executor:
    def __init__(self):
        self._is_executor = True
        self._is_complete = False
        self.result = None

    def is_executor(self):
        return self._is_executor

    def is_complete(self):
        return self._is_complete

    def execute(self):
        raise Exception("Method is not implemented.")

    def _set_result(self, **data):
        self.result = self.Result(**data)

    class Result:
        def __init__(self, **data):
            self.__dict__.update(data)



# Terraform Test Framework
# https://github.com/tf2project/tf2project

from json import loads as load_json
from os import environ, getcwd
from os.path import abspath, exists
from subprocess import run as run_subprocess

from .loader import TerraformLoader

TF_CLI_CHDIR = environ.get("TF_CLI_CHDIR", getcwd())


class TerraformLocalFileLoader(TerraformLoader):
    def __init__(self, file_path):
        super().__init__()
        self._data_path = abspath(file_path)

    def load(self):
        if exists(self._data_path) is False:
            raise Exception(f"File '{ self._data_path }' is not found.")
        result = run_subprocess(
            f"terraform -chdir={ TF_CLI_CHDIR } show -json { self._data_path }",
            shell=True,
            capture_output=True,
        )
        try:
            return load_json(result.stdout)
        except:
            raise Exception(f"File '{ self._data_path }' could not be loaded.")


class TerraformPlanLoader(TerraformLocalFileLoader):
    def __init__(self, file_path=None):
        file_path = "./terraform.tfplan" if file_path is None else file_path
        super().__init__(file_path)
        self._loader_type = "planloader"


class TerraformStateLoader(TerraformLocalFileLoader):
    def __init__(self, file_path=None):
        file_path = "./terraform.tfstate" if file_path is None else file_path
        super().__init__(file_path)
        self._loader_type = "stateloader"



# Terraform Test Framework
# https://github.com/tf2project/tf2project


class TerraformLoader:
    def __init__(self):
        self._is_loader = True
        self._loader_type = None
        self._data_path = None

    def load(self):
        raise Exception("Method is not implemented.")

    def is_loader(self):
        return self._is_loader

    def get_loader_type(self):
        return self._loader_type

    def get_data_path(self):
        return self._data_path



# Terraform Test Framework
# https://github.com/tf2project/tf2project

from sys import exit
from time import perf_counter

from .cli import print_run_footer, print_run_header, print_test_result
from .terraform import Terraform
from .test import Test


class Tf2:
    def __init__(
        self,
        terraform_instance=None,
        ignore_errors=False,
        silent_mode=False,
        create_result_file=True,
    ):
        if terraform_instance is None:
            self._terraform_instance = Terraform()
        else:
            self._terraform_instance = terraform_instance
        self._ignore_errors = ignore_errors
        self._silent_mode = silent_mode
        self._create_result_file = create_result_file
        self._fetch_objects()
        self._tests = []

    def _fetch_objects(self):
        for object_name in ["resources", "data", "modules", "outputs"]:
            if hasattr(self._terraform_instance, object_name):
                setattr(
                    self,
                    object_name,
                    getattr(self._terraform_instance, object_name),
                )

    def _find_testable_object(self, object_name):
        current_object = self
        target_object_tree = object_name.split(".")
        for target_object_name in target_object_tree:
            if hasattr(current_object, target_object_name) is False:
                raise Exception(f"Object '{ object_name }' is not available.")
            target_object = getattr(current_object, target_object_name)
            if target_object.is_testable() is False:
                raise Exception(f"Object '{ object_name }' is not testable.")
            current_object = target_object
        return current_object

    def add_test(self, object_name, test_func, ignore_errors=None):
        self._tests.append(
            Test(
                object_name=object_name,
                object_instance=self._find_testable_object(object_name),
                test_func=test_func,
                ignore_errors=self._ignore_errors
                if ignore_errors is None
                else ignore_errors,
            )
        )

    def test(self, object_name, ignore_errors=None):
        test_wrapper = lambda test_func: self.add_test(
            object_name, test_func, ignore_errors
        )
        return test_wrapper

    def run(self):
        if self._silent_mode is False:
            print_run_header(
                self._terraform_instance.get_data()["terraform_version"],
                self._terraform_instance.get_data()["format_version"],
                self._terraform_instance.get_loader_instance().get_loader_type(),
                self._terraform_instance.get_loader_instance().get_data_path(),
                len(self._tests),
            )
        total_passed_tests = total_failed_tests = 0
        overall_result = True
        start_time = perf_counter()
        for test in self._tests:
            current_test_result = True
            try:
                if "self" in test.test_func.__code__.co_varnames:
                    test.test_func(test.object_instance)
                else:
                    test.test_func()
            except:
                current_test_result = False
            if current_test_result is True:
                total_passed_tests += 1
            else:
                total_failed_tests += 1
                if test.ignore_errors is False:
                    overall_result = False
            if self._silent_mode is False:
                print_test_result(
                    test.object_name,
                    test.test_func.__name__,
                    current_test_result,
                )
            if current_test_result is False and test.ignore_errors is False:
                break
        total_time = perf_counter() - start_time
        if self._silent_mode is False:
            print_run_footer(
                overall_result,
                total_failed_tests,
                total_passed_tests,
                total_time,
            )
        if self._create_result_file is True:
            with open(".tf2result", "w") as f:
                f.write("passed") if overall_result else f.write("failed")
                f.write("\n")
        exit(not overall_result)



# Terraform Test Framework
# https://github.com/tf2project/tf2project

from importlib.metadata import version as package_version
from os import get_terminal_size, getcwd
from platform import python_version
from platform import system as system_version

RUN_HEADER = """\033[1m{0}\033[0m
   \033[92m__________   ________    _______
  |\___   ___\ |\  _____\  /  ___  \\
  \|___ \  \_| \ \  \__/  /__/|_/  /|\033[0m
       \ \  \   \ \   __\ |__|//  / /
        \ \  \   \ \  \_|     /  /_/__
         \033[91m\ \__\   \ \__\     |\________\\ Dedicated to Iran and Iranians
          \|__|    \|__|     \|________| Terraform Test Framework\033[0m

platform {1}, python {2}
tf2 {3}, terraform {4}, json {5}, type {6}
rootdir: {7}
datapath: {8}
\033[1mcollected {9} {10}\033[0m
"""

RUN_FOOTER = """\033[1m{0}{1}\033[0m"""

OUTPUT_STRING = """{0} in {1}s"""

FULLWIDTH_LINE = """{0} {1} {2}"""

TEST_RESULT = """object_name: \033[1m{0}\033[0m
\u2799 test_func: \033[1m{1}\033[0m {2} \033[1m{3}\033[0m
"""


def create_fullwidth_line(input_string):
    try:
        available_columns = get_terminal_size().columns - len(input_string)
    except:
        available_columns = 80 - len(input_string)
    available_columns = 4 if available_columns < 4 else available_columns
    left_available_columns = available_columns // 2
    right_available_columns = available_columns // 2 + available_columns % 2
    return FULLWIDTH_LINE.format(
        "=" * (left_available_columns - 1),
        input_string,
        "=" * (right_available_columns - 1),
    )


def print_test_result(object_name, test_func, test_result):
    print(
        TEST_RESULT.format(
            object_name,
            test_func,
            "\u2705" if test_result else "\u274c",
            "\033[92mPASSED" if test_result else "\033[91mFAILED",
        )
    )


def print_run_header(
    terraform_version,
    format_version,
    loader_type,
    terraform_data_path,
    total_tests,
):
    print(
        RUN_HEADER.format(
            create_fullwidth_line("tf2"),
            system_version().lower(),
            python_version(),
            package_version("tf2project"),
            terraform_version,
            format_version,
            loader_type,
            getcwd(),
            terraform_data_path,
            total_tests,
            "tests" if total_tests > 1 else "test",
        )
    )


def print_run_footer(
    overall_result,
    total_failed_tests,
    total_passed_tests,
    total_time,
):
    output_contents = []
    if total_failed_tests:
        output_contents.extend([str(total_failed_tests), "failed"])
    if total_passed_tests:
        output_contents.extend([str(total_passed_tests), "passed"])
    if not total_failed_tests and not total_passed_tests:
        output_contents.append("nothing")
    print(
        RUN_FOOTER.format(
            "\033[92m" if overall_result else "\033[91m",
            create_fullwidth_line(
                OUTPUT_STRING.format(
                    " ".join((output_contents)),
                    round(total_time, 2),
                ),
            ),
        )
    )



# Terraform Test Framework
# https://github.com/tf2project/tf2project

from os import environ

from setuptools import setup

if environ.get("ENV", "production") == "development":
    with open("requirements.txt") as f:
        requirements = f.readlines()
else:
    requirements = []

with open("README.md") as f:
    readme = f.read()

setup(
    long_description=readme,
    install_requires=requirements,
)



# Terraform Test Framework
# https://github.com/tf2project/tf2project

from re import match, sub


def get_output_type(value):
    return match("<class '(.*)'>", str(type(value)))[1]


def get_attr_name(value):
    for pattern in ["\W", "_+"]:
        value = sub(pattern, "_", value)
    return value



# Terraform Test Framework
# https://github.com/tf2project/tf2project


class Test:
    def __init__(
        self,
        object_name,
        object_instance,
        test_func,
        ignore_errors,
    ):
        self.object_name = object_name
        self.object_instance = object_instance
        self.test_func = test_func
        self.ignore_errors = ignore_errors

    def get(self, name):
        return getattr(self, name, None)

    def set(self, name, value):
        setattr(self, name, value)



# Terraform Test Framework
# https://github.com/tf2project/tf2project

from .util import get_attr_name


class TerraformObject:
    def __init__(self, is_testable=True):
        self._is_testable = is_testable

    def is_testable(self):
        return self._is_testable


class TerraformObjectParser(TerraformObject):
    def __init__(self, data, is_testable=True):
        super().__init__(is_testable)
        for _key, value in data.items():
            key = get_attr_name(_key)
            if isinstance(value, dict) is True:
                setattr(self, key, TerraformObjectParser(value, False))
            elif isinstance(value, (list, tuple)) is True:
                items = [
                    TerraformObjectParser(item, False)
                    if isinstance(item, dict)
                    else item
                    for item in value
                ]
                setattr(self, key, items)
            else:
                setattr(self, key, value)
